'use strict';

var Concur = require('Concur')
var getFormData = require('get-form-data')
var copy = require('isomorph/copy')
var $__0=  require('isomorph/format'),formatObj=$__0.formatObj
var is = require('isomorph/is')
var object = require('isomorph/object')

var constants = require('./constants')

var BoundField = require('./BoundField')
var DeclarativeFieldsMeta = require('./DeclarativeFieldsMeta')
var ErrorList = require('./ErrorList')
var ErrorObject = require('./ErrorObject')
var FileField = require('./FileField')
var MultipleFileField = require('./MultipleFileField')

var $__1=  require('validators'),ValidationError=$__1.ValidationError
var $__2=      require('./util'),cancellable=$__2.cancellable,debounce=$__2.debounce,info=$__2.info,warning=$__2.warning,normaliseValidation=$__2.normaliseValidation

function noop() {}
var sentinel = {}

var NON_FIELD_ERRORS = constants.NON_FIELD_ERRORS

/**
 * Checks if a field's view of raw input data (via its Widget) has changed.
 */
function fieldDataHasChanged(previous, current) {
  if (is.Array(previous) && is.Array(current)) {
    if (previous.length != current.length) { return true }
    for (var i = 0, l = previous.length; i < l; i++) {
      if (previous[i] != current[i]) { return true }
    }
    return false
  }
  return previous != current
}

if ('production' !== process.env.NODE_ENV) {
  var warnedOnImpliedValidateAuto = false
}

/**
 * A collection of Fields that knows how to validate and display itself.
 * @constructor
 * @param {Object.<string, *>} kwargs form options.
 */
var Form = Concur.extend({
  __meta__: DeclarativeFieldsMeta,

  prefixFormat: '{prefix}-{name}',

  constructor: function Form(kwargs) {
    // TODO Perform PropType checks on kwargs in development mode
    kwargs = object.extend({
      data: null, files: null, autoId: 'id_{name}', prefix: null,
      initial: null, errorConstructor: ErrorList, labelSuffix: ':',
      emptyPermitted: false, validation: null, controlled: false,
      onChange: null, errors: null
    }, kwargs)
    this.isInitialRender = (kwargs.data == null && kwargs.files == null)
    this.data = kwargs.data || {}
    this.files = kwargs.files || {}
    this.autoId = kwargs.autoId
    this.prefix = kwargs.prefix
    this.initial = kwargs.initial || {}
    this.cleanedData = {}
    this.errorConstructor = kwargs.errorConstructor
    this.labelSuffix = kwargs.labelSuffix
    this.emptyPermitted = kwargs.emptyPermitted
    this.controlled = kwargs.controlled
    this.onChange = kwargs.onChange

    // Auto validation is implied when onChange is passed
    if (is.Function(kwargs.onChange)) {
      if ('production' !== process.env.NODE_ENV) {
        if (!warnedOnImpliedValidateAuto && kwargs.validation === 'auto') {
          info('Passing onChange to a Form or FormSet constructor also ' +
               "implies validation: 'auto' by default - you don't have " +
               'to set it manually.')
          warnedOnImpliedValidateAuto = true
        }
      }
      if (kwargs.validation == null) {
        kwargs.validation = 'auto'
      }
    }
    this.validation = normaliseValidation(kwargs.validation || 'manual')

    this._errors = kwargs.errors

    // Cancellable debounced functions for delayed event validation
    this._pendingEventValidation = {}
    // Input data as it was last time validation was performed on a field
    this._lastValidatedData = {}
    // Cached result of the last call to hasChanged()
    this._lastHasChanged = null

    // Lookup for names of fields pending validation
    this._pendingValidation = {}
    // Cancellable callbacks for pending async validation
    this._pendingAsyncValidation = {}
    // Lookup for names of fields pending validation which clean() depends on
    this._runCleanAfter = {}
    // Callback to be run the next time validation finishes
    this._onValidate = null

    // The baseFields attribute is the *prototype-wide* definition of fields.
    // Because a particular *instance* might want to alter this.fields, we
    // create this.fields here by deep copying baseFields. Instances should
    // always modify this.fields; they should not modify baseFields.
    this.fields = copy.deepCopy(this.baseFields)

    if ('production' !== process.env.NODE_ENV) {
      // Now that form.fields exists, we can check if there's any configuration
      // which *needs* onChange on the form or its fields.
      if (!is.Function(kwargs.onChange) && this._needsOnChange()) {
        warning("You didn't provide an onChange callback for a " +
                this._formName() + ' which has controlled fields. This ' +
                'will result in read-only fields.')
      }
    }

    // Copy initial values to the data object, as it represents form input -
    // literally so in the case of controlled components once we start taking
    // some data and isInitialRender flips to false.
    if (this.isInitialRender) {
      this._copyInitialToData()
    }
  }
})

// XXX Don't alter form extension arguments - fix this in Concur
var _extend = Form.extend
Form.extend = function(prototypeProps, constructorProps) {
  return _extend.call(this, object.extend({}, prototypeProps), constructorProps)
}

/**
 * Calls the onChange function if it's been provided. This method will be called
 * every time the form makes a change to its state which requires redisplay.
 */
Form.prototype._stateChanged = function() {
  if (typeof this.onChange == 'function') {
    this.onChange()
  }
}

/**
 * Copies initial data to the input data object, as it represents form input -
 * when using controlled components once we start taking some data,
 * isInitialRender flips to false and this.data is used for rendering widgets.
 */
Form.prototype._copyInitialToData = function() {
  var initialData = object.extend(this._fieldInitialData(), this.initial)
  var initialFieldNames = Object.keys(initialData)
  for (var i = 0, l = initialFieldNames.length; i < l; i++) {
    var fieldName = initialFieldNames[i]
    if (typeof this.fields[fieldName] == 'undefined') { continue }
    // Don't copy initial to input data for fields which can't have the
    // initial data set as their current value.
    if (!this.fields[fieldName].widget.isValueSettable) { continue }
    this.data[this.addPrefix(fieldName)] = initialData[fieldName]
  }
}

/**
 * Gets initial data configured in this form's fields.
 * @return {Object.<string,*>}
 */
Form.prototype._fieldInitialData = function() {
  var fieldInitial = {}
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l; i++) {
    var fieldName = fieldNames[i]
    var initial = this.fields[fieldName].initial
    if (initial !== null) {
      fieldInitial[fieldName] = initial
    }
  }
  return fieldInitial
}

/**
 * Tries to construct a display name for the form for display in messages.
 * @return {string}
 */
Form.prototype._formName = function() {
  var name = this.displayName || this.constructor.name
  return (name ? "'" + name + "'" : 'Form')
}

/**
 * @return {boolean} true if the form or any of its fields are configured to
 *   generate controlled components.
 */
Form.prototype._needsOnChange = function() {
  if (this.controlled === true) {
    return true
  }
  var names = Object.keys(this.fields)
  for (var i = 0, l = names.length; i < l; i++) {
    if (this.fields[names[i]].controlled === true) {
      return true
    }
  }
  return false
}

// ============================================================== Validation ===

/**
 * Validates the form from scratch. If a <form> is given, data from it will be
 * set on this form first. Otherwise, validation will be done with this form's
 * current input data.
 * @param {(ReactElement|HTMLFormElement)=} form the <form> containing this
 * form's rendered widgets - this can be a React <form> component or a real
 * <form> DOM node.
 * @param {function(err, isValid, cleanedData)=} cb callback for asynchronous
 *   validation.
 * @return {boolean|undefined} true if the form only has synchronous validation
 *   and is valid.
 * @throws if the form has asynchronous validation and a callback is not
 *   provided.
 */
Form.prototype.validate = function(form, cb) {
  this._cancelPendingOperations()
  if (is.Function(form)) {
    cb = form
    form = null
  }
  if (form) {
    if (typeof form.getDOMNode == 'function') {
      form = form.getDOMNode()
    }
    this.data = getFormData(form)
  }
  return (this.isAsync() ? this._validateAsync(cb) : this._validateSync())
}

Form.prototype._validateAsync = function(cb) {
  if (!is.Function(cb)) {
    throw new Error(
      'You must provide a callback to validate() when a form has ' +
      'asynchronous validation.'
    )
  }
  if (this.isInitialRender) {
    this.isInitialRender = false
  }
  this._onValidate = cb
  this.fullClean()
  // Display async progress indicators
  this._stateChanged()
}

Form.prototype._validateSync = function() {
  if (this.isInitialRender) {
    this.isInitialRender = false
  }
  this.fullClean()
  // Display changes to valid/invalid state
  this._stateChanged()
  return this.isValid()
}

/**
 * Cleans data for all fields and triggers cross-form cleaning.
 */
Form.prototype.fullClean = function() {
  this._errors = new ErrorObject()
  if (this.isInitialRender) {
    return // Stop further processing
  }

  this.cleanedData = {}

  // If the form is permitted to be empty, and none of the form data has
  // changed from the initial data, short circuit any validation.
  if (this.emptyPermitted && !this.hasChanged()) {
    this._finishedValidation(null)
    return
  }

  this._cleanFields()
}

/**
 * Cleans data for the given field names and triggers cross-form cleaning in
 * case any cleanedData it uses has changed.
 * @param {Array.<string>} fields field names.
 */
Form.prototype.partialClean = function(fields) {
  this._removeErrors(fields)

  // If the form is permitted to be empty, and none of the form data has
  // changed from the initial data, short circuit any validation.
  if (this.emptyPermitted && !this.hasChanged()) {
    if (this._errors.isPopulated()) {
      this._errors = ErrorObject()
    }
    return
  }

  this._preCleanFields(fields)
  for (var i = 0, l = fields.length; i < l; i++) {
    this._cleanField(fields[i])
  }
}

/**
 * Validates and cleans every field in the form.
 */
Form.prototype._cleanFields = function() {
  var fieldNames = Object.keys(this.fields)
  this._preCleanFields(fieldNames)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    this._cleanField(fieldNames[i])
  }
}

/**
 * Sets up pending validation state prior to cleaning fields and configures
 * cross-field cleaning to run after its dependent fields have been cleaned, or
 * after all fields have been cleaned if dependencies have not been configured.
 * @param {Array.<string>} fieldNames fields which are about to be cleaned.
 */
Form.prototype._preCleanFields = function(fieldNames) {
  // Add all field names to those pending validation
  object.extend(this._pendingValidation, object.lookup(fieldNames))

  // Add appropriate field names to determine when to run cross-field cleaning
  var i, l
  if (typeof this.clean.fields != 'undefined') {
    for (i = 0, l = fieldNames.length; i < l; i++) {
      if (this.clean.fields[fieldNames[i]]) {
        this._runCleanAfter[fieldNames[i]] = true
      }
    }
  }
  else {
    // Ignore any invalid field names given
    for (i = 0, l = fieldNames.length; i < l; i++) {
      if (this.fields[fieldNames[i]]) {
        this._runCleanAfter[fieldNames[i]] = true
      }
    }
  }
}

/**
 * Validates and cleans the named field and runs any custom validation function
 * that's been provided for it.
 * @param {string} name the name of a form field.
 */
Form.prototype._cleanField = function(name) {
  if (!object.hasOwn(this.fields, name)) {
    throw new Error(this._formName() + " has no field named '" + name + "'")
  }

  var field = this.fields[name]
  // valueFromData() gets the data from the data objects.
  // Each widget type knows how to retrieve its own data, because some widgets
  // split data over several HTML fields.
  var value = field.widget.valueFromData(this.data, this.files,
                                         this.addPrefix(name))
  var async = false
  var error = null

  try {
    if (field instanceof FileField) {
      var initial = object.get(this.initial, name, field.initial)
      value = field.clean(value, initial)
    }
    else {
      value = field.clean(value)
    }
    this.cleanedData[name] = value
    var customClean = this._getCustomClean(name)
    if (is.Function(customClean)) {
      async = this._runCustomClean(name, customClean)
    }
  }
  catch (e) {
    if (e instanceof ValidationError) {
      this.addError(name, e)
    }
    else {
      error = e
    }
  }

  if (!async) {
    this._fieldCleaned(name, error)
  }
}

/**
 * Gets the custom cleaning method for a field. These can be named clean<Name>
 * or clean_<name>.
 * @param {string} fieldName
 * @return {function|undefined}
 */
Form.prototype._getCustomClean = function(fieldName) {
  return (this['clean' + fieldName.charAt(0).toUpperCase() + fieldName.substr(1)] ||
          this['clean_' + fieldName])
}

/**
 * Calls a custom cleaning method, expecting synchronous or asynchronous
 * behaviour, depending on its arity.
 * @param {string} fieldName a field name.
 * @param {(function()|function(function(Error, string, string|ValidationError)))} customClean
 *   the custom cleaning method for the field.
 * @return {boolean} true if cleaning is running asynchronously, false if it just
 *   ran synchronously.
 */
Form.prototype._runCustomClean = function(fieldName, customClean) {
  // Check arity to see if we have a callback in the function signature
  if (customClean.length === 0) {
    // Synchronous processing only expected
    customClean.call(this)
    return false
  }

  // If custom validation is async and there's one pending, prevent its
  // callback from doing anything.
  if (typeof this._pendingAsyncValidation[fieldName] != 'undefined') {
    object.pop(this._pendingAsyncValidation, fieldName).cancel()
  }
  // Set up callback for async processing - the argument for addError()
  // should be passed via the callback as calling it directly prevents us
  // from completely ignoring the callback if validation fires again.
  var callback = function(err, validationError) {
    if (validationError) {
      this.addError(fieldName == NON_FIELD_ERRORS ? null : fieldName, validationError)
    }
    this._fieldCleaned(fieldName, err)
    this._stateChanged()
  }.bind(this)
  var cancellableCallback = cancellable(callback)

  // An explicit return value of false indicates that async processing is
  // being skipped (e.g. because sync checks in the method failed first)
  var returnValue = customClean.call(this, cancellableCallback)
  if (returnValue !== false) {
    // Async processing is happening! Make the callback cancellable and
    // hook up any custom onCancel handling provided.
    if (returnValue && typeof returnValue.onCancel == 'function') {
      callback.onCancel = returnValue.onCancel
    }
    this._pendingAsyncValidation[fieldName] = cancellableCallback
    return true
  }
}

/**
 * Callback for completion of field cleaning. Triggers further field cleaning or
 * signals the end of validation, as necessary.
 * @param {string} fieldName
 * @param {Error=} err an error caught while cleaning the field.
 */
Form.prototype._fieldCleaned = function(fieldName, err) {
  var wasPending = delete this._pendingValidation[fieldName]
  if (this._pendingAsyncValidation[fieldName]) {
    delete this._pendingAsyncValidation[fieldName]
  }

  if (err) {
    if ("production" !== process.env.NODE_ENV) {
      console.error('Error cleaning ' + this._formName() + '.' + fieldName +
                    ':' + err.message)
    }
    // Stop tracking validation progress on error, and don't call clean()
    this._pendingValidation = {}
    this._runCleanAfter = {}
    this._finishedValidation(err)
    return
  }

  // Run clean() if this was the last field it was waiting for
  if (this._runCleanAfter[fieldName]) {
    delete this._runCleanAfter[fieldName]
    if (is.Empty(this._runCleanAfter)) {
      this._cleanForm()
      return
    }
  }

  // Signal the end of validation if this was the last field we were waiting for
  if (wasPending && is.Empty(this._pendingValidation)) {
    this._finishedValidation(null)
  }
}

/**
 * Hook for doing any extra form-wide cleaning after each Field has been cleaned.
 * Any ValidationError thrown by synchronous validation in this method will not
 * be associated with a particular field; it will have a special-case association
 * with the field named '__all__'.
 * @param {function(Error, string, string|ValidationError)=} cb a callback to signal the
 *   end of asynchronous validation.
 */
Form.prototype.clean = noop

/**
 * Calls the clean() hook.
 */
Form.prototype._cleanForm = function() {
  var async = false
  var error = null
  try {
    if (this.clean !== noop) {
      async = this._runCustomClean(NON_FIELD_ERRORS, this.clean)
    }
  }
  catch (e) {
    if (e instanceof ValidationError) {
      this.addError(null, e)
    }
    else {
      error = e
    }
  }

  if (!async) {
    this._fieldCleaned(NON_FIELD_ERRORS, error)
  }
}

Form.prototype._finishedValidation = function(err) {
  if (!this.isAsync()) {
    if (err) {
      throw err
    }
    // Synchronous form validation results will be returned via the original
    // call which triggered validation.
    return
  }
  if (is.Function(this._onValidate)) {
    var callback = this._onValidate
    this._onValidate = null
    if (err) {
      return callback(err)
    }
    var isValid = this.isValid()
    callback(null, isValid, isValid ? this.cleanedData : null)
  }
}

/**
 * Cancels any pending field validations and async validations.
 */
Form.prototype._cancelPendingOperations = function() {
  Object.keys(this._pendingEventValidation).forEach(function(field) {
    object.pop(this._pendingEventValidation, field).cancel()
  }.bind(this))
  Object.keys(this._pendingAsyncValidation).forEach(function(field) {
    object.pop(this._pendingAsyncValidation, field).cancel()
  }.bind(this))
}

// ========================================================== Event Handling ===

/**
 * Handles validating the field which is the target of the given event based
 * on its validation config. This will be hooked up to the appropriate event
 * as per the field's validation config.
 * @param {Object} validation the field's validation config for the event.
 * @param {SyntheticEvent} e the event being handled.
 */
Form.prototype._handleFieldEvent = function(validation, e) {
  // Update form.data with the current value of the field which is the target of
  // the event.
  var htmlName = e.target.name
  var fieldName = this.removePrefix(e.target.getAttribute('data-newforms-field') || htmlName)
  var field = this.fields[fieldName]
  var targetData = getFormData.getNamedFormElementData(e.target.form, htmlName)
  this.data[htmlName] = targetData
  if (field instanceof FileField && 'files' in e.target) {
    var files = e.target.files
    this.files[htmlName] = (field instanceof MultipleFileField
                            ? Array.prototype.slice.call(files)
                            : files[0])
  }
  if (this.isInitialRender) {
    this.isInitialRender = false
  }
  if (this.controlled || field.controlled) {
    this._stateChanged()
  }

  // Bail out early if the event is only being handled to update the field's data
  if (validation.validate === false) { return }

  var validate = false

  // Special cases for onBlur, as it ends a user's interaction with a text input
  if (validation.event == 'onBlur') {
    // If there is any pending validation, trigger it immediately
    if (typeof this._pendingEventValidation[fieldName] != 'undefined') {
      this._pendingEventValidation[fieldName].trigger()
      return
    }
    // Always validate if the field is required and the input which was blurred
    // was empty (some fields have multiple inputs).
    validate = (field.required && field.isEmptyValue(targetData))
  }

  // Always validate if this is the first time the field has been interacted
  // with.
  if (!validate) {
    var lastValidatedData = object.get(this._lastValidatedData, fieldName, sentinel)
    validate = (lastValidatedData === sentinel)
  }

  // Otherwise, validate if data has changed since validation was last performed
  // - this prevents displayed validation errors being cleared unnecessarily.
  if (!validate) {
    var fieldData = field.widget.valueFromData(this.data, this.files, this.addPrefix(fieldName))
    validate = fieldDataHasChanged(lastValidatedData, fieldData)
  }

  // Cancel any pending validation as it's no longer needed - this can happen
  // if the user edits a field with debounced validation and it ends up back
  // at its original value before validation is triggered.
  if (!validate && typeof this._pendingEventValidation[fieldName] != 'undefined') {
    object.pop(this._pendingEventValidation, fieldName).cancel()
  }

  // If we don't need to validate, we're done handling the event
  if (!validate) { return }

  if (validation.delay) {
    this._delayedFieldValidation(fieldName, validation.delay)
  }
  else {
    this._immediateFieldValidation(fieldName)
  }
}

/**
 * Sets up delayed validation of a field with a debounced function and calls it,
 * or just calls the function again if it already exists, to reset the delay.
 * @param {string} fieldName
 * @param {number} delay delay time in ms.
 */
Form.prototype._delayedFieldValidation = function(fieldName, delay) {
  if (typeof this._pendingEventValidation[fieldName] == 'undefined') {
    this._pendingEventValidation[fieldName] = debounce(function() {
      delete this._pendingEventValidation[fieldName]
      this._immediateFieldValidation(fieldName)
    }.bind(this), delay)
  }
  this._pendingEventValidation[fieldName]()
}

/**
 * Validates a field and notifies the React component that state has changed.
 * @param {string} fieldName
 */
Form.prototype._immediateFieldValidation = function(fieldName) {
  // Remove and cancel any pending validation for the field to avoid doubling up
  // when both delayed and immediate validation are configured.
  if (typeof this._pendingEventValidation[fieldName] != 'undefined') {
    object.pop(this._pendingEventValidation, fieldName).cancel()
  }
  this._lastValidatedData[fieldName] =
      this.fields[fieldName].widget.valueFromData(this.data, this.files,
                                                  this.addPrefix(fieldName))
  this.partialClean([fieldName])
  this._stateChanged()
}

// ============================================================== Mutability ===

/**
 * Resets a form data back to its initial state, optionally providing new initial
 * data.
 * @param {Object.<string, *>=} newInitial new initial data for the form.
 */
Form.prototype.reset = function(newInitial) {
  this._cancelPendingOperations()

  if (typeof newInitial != 'undefined') {
    this.initial = newInitial
  }

  this.data = {}
  this.cleanedData = {}
  this.isInitialRender = true

  this._errors = null
  this._lastHasChanged = null
  this._pendingValidation = {}
  this._runCleanAfter = {}
  this._lastValidatedData = {}
  this._onValidate = null

  this._copyInitialToData()
  this._stateChanged()
}

/**
 * Sets the form's entire input data, also triggering validation by default.
 * @param {object.<string,*>} data new input data for the form.
 * @param {object.<string,boolean>} kwargs data setting options.
 * @return {boolean|undefined} if data setting options indicate the new data
 *   should be validated and the form does not have asynchronous validation
 *   configured: true if the new data is valid.
 */
Form.prototype.setData = function(data, kwargs) {
  kwargs = object.extend({
    prefixed: false, validate: true, _triggerStateChange: true
  }, kwargs)

  this.data = (kwargs.prefixed ? data : this._prefixData(data))

  if (this.isInitialRender) {
    this.isInitialRender = false
  }
  if (kwargs.validate) {
    this._errors = null
    // This call ultimately triggers a fullClean() because _errors is null
    var isValid = this.isValid()
  }
  else {
    // Prevent validation being triggered if errors() is accessed during render
    this._errors = new ErrorObject()
  }

  if (kwargs._triggerStateChange) {
    this._stateChanged()
  }

  if (kwargs.validate && !this.isAsync()) {
    return isValid
  }
}

/**
 * Sets the form's entire input data wth data extracted from a <form>, which
 * will be prefixed, if prefixes are being used.
 * @param {Object.<strong, *>} formData
 * @param {Object.<string, boolean>} kwargs setData options.
 */
Form.prototype.setFormData = function(formData, kwargs) {
  return this.setData(formData, object.extend(kwargs || {}, {prefixed: true}))
}

/**
 * Updates some of the form's input data, optionally triggering validation of
 * updated fields and form-wide cleaning, or clears existing errors from the
 * updated fields.
 * @param {Object.<string, *>} data updated input data for the form.
 * @param {Object.<string, boolean>} kwargs update options.
 */
Form.prototype.updateData = function(data, kwargs) {
  kwargs = object.extend({
    prefixed: false, validate: true, clearValidation: true
  }, kwargs)

  object.extend(this.data, (kwargs.prefixed ? data : this._prefixData(data)))
  if (this.isInitialRender) {
    this.isInitialRender = false
  }

  var fields = Object.keys(data)
  if (kwargs.prefixed) {
    fields = fields.map(this.removePrefix.bind(this))
  }

  if (kwargs.validate) {
    this.partialClean(fields)
  }
  else if (kwargs.clearValidation) {
    this._removeErrors(fields)
    this._removeCleanedData(fields)
    this._cleanForm()
  }

  this._stateChanged()
}

/**
 * Removes any cleanedData present for the given form fields.
 * @param {Array.<string>} fields field names.
 */
Form.prototype._removeCleanedData = function(fields) {
  for (var i = 0, l = fields.length; i < l; i++) {
    delete this.cleanedData[fields[i]]
  }
}

// ============================================================= BoundFields ===

/**
 * Creates a BoundField for the field with the given name.
 * @param {string} name a field name.
 * @return {BoundField} a BoundField for the field.
 */
Form.prototype.boundField = function(name) {
  if (!object.hasOwn(this.fields, name)) {
    throw new Error(this._formName() + " does not have a '" + name + "' field.")
  }
  return new BoundField(this, this.fields[name], name)
}

/**
 * Creates a BoundField for each field in the form, in the order in which the
 * fields were created.
 * @param {function(Field, string)=} test if provided, this function will be
 *   called with field and name arguments - BoundFields will only be generated
 *   for fields for which true is returned.
 * @return {Array.<BoundField>} a list of BoundField objects.
 */
Form.prototype.boundFields = function(test) {
  var bfs = []
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    var fieldName = fieldNames[i]
    if (!test || test(this.fields[fieldName], fieldName)) {
      bfs.push(new BoundField(this, this.fields[fieldName], fieldName))
    }
  }
  return bfs
}

/**
 * Like boundFields(), but returns a name -> BoundField object instead.
 * @return {Object.<string, BoundField>}
 */
Form.prototype.boundFieldsObj = function() {
  var bfs = {}
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    var fieldName = fieldNames[i]
    bfs[fieldName] = new BoundField(this, this.fields[fieldName], fieldName)
  }
  return bfs
}

/**
 * Returns a list of all the BoundField objects that correspond to hidden
 * fields. Useful for manual form layout.
 * @return {Array.<BoundField>}
 */
Form.prototype.hiddenFields = function() {
  return this.boundFields(function(field) {
    return field.widget.isHidden
  })
}

/**
 * Returns a list of BoundField objects that do not correspond to hidden fields.
 * The opposite of the hiddenFields() method.
 * @return {Array.<BoundField>}
 */
Form.prototype.visibleFields = function() {
  return this.boundFields(function(field) {
    return !field.widget.isHidden
  })
}

// ================================================================== Errors ===

/**
 * Updates the content of this._errors.
 * The field argument is the name of the field to which the errors should be
 * added. If its value is null the errors will be treated as NON_FIELD_ERRORS.
 * The error argument can be a single error, a list of errors, or an object that
 * maps field names to lists of errors. What we define as an "error" can be
 * either a simple string or an instance of ValidationError with its message
 * attribute set and what we define as list or object can be an actual list or
 * object or an instance of ValidationError with its errorList or errorObj
 * property set.
 * If error is an object, the field argument *must* be null and errors will be
 * added to the fields that correspond to the properties of the object.
 * @param {?string} field the name of a form field.
 * @param {(string|ValidationError|Array.<(string|ValidationError)>|Object<string,(string|ValidationError|Array.<(string|ValidationError)>))} error
 */
Form.prototype.addError = function(field, error) {
  if (!(error instanceof ValidationError)) {
    // Normalise to ValidationError and let its constructor do the hard work of
    // making sense of the input.
    error = ValidationError(error)
  }

  if (object.hasOwn(error, 'errorObj')) {
    if (field !== null) {
      throw new Error("The 'field' argument to form.addError() must be null when " +
                      "the 'error' argument contains errors for multiple fields.")
    }
    error = error.errorObj
  }
  else {
    var errorList = error.errorList
    error = {}
    error[field || NON_FIELD_ERRORS] = errorList
  }

  var fields = Object.keys(error)
  for (var i = 0, l = fields.length; i < l; i++) {
    field = fields[i]
    errorList = error[field]
    if (!this._errors.hasField(field)) {
      if (field !== NON_FIELD_ERRORS && !object.hasOwn(this.fields, field)) {
        throw new Error(this._formName() + " has no field named '" + field + "'")
      }
      this._errors.set(field, new this.errorConstructor())
    }
    else {
      // Filter out any error messages which are duplicates of existing
      // messages. This can happen if onChange validation which uses addError()
      // is fired repeatedly and is adding an error message to a field other
      // then the one being changed.
      var messageLookup = object.lookup(this._errors.get(field).messages())
      var newMessages = ErrorList(errorList).messages()
      for (var j = errorList.length - 1; j >= 0; j--) {
        if (messageLookup[newMessages[j]]) {
          errorList.splice(j, 1)
        }
      }
    }

    if (errorList.length > 0) {
      this._errors.get(field).extend(errorList)
    }

    if (object.hasOwn(this.cleanedData, field)) {
      delete this.cleanedData[field]
    }
  }
}

/**
 * Getter for errors, which first cleans the form if there are no errors
 * defined yet.
 * @param {string=} name if given, errors for this field name will be returned
 *   instead of the full error object.
 * @return {ErrorObject|ErrorList} form or field errors
 */
Form.prototype.errors = function(name) {
  if (this._errors == null) {
    this.fullClean()
  }
  if (name) {
    return this._errors.get(name)
  }
  return this._errors
}

/**
 * @return {ErrorObject} errors that aren't associated with a particular field -
 *   i.e., errors generated by clean(). Will be empty if there are none.
 */
Form.prototype.nonFieldErrors = function() {
  return (this.errors(NON_FIELD_ERRORS) || new this.errorConstructor())
}

/**
 * @param {ErrorObject} errors
 */
Form.prototype.setErrors = function(errors) {
  this._errors = errors
  this._stateChanged()
}

/**
 * Removes any validation errors present for the given form fields. If validation
 * has not been performed yet, initialises the errors object.
 * @param {Array.<string>} fields field names.
 */
Form.prototype._removeErrors = function(fields) {
  if (this._errors == null) {
    this._errors = ErrorObject()
  }
  else {
    // TODO use clean.fields if available
    this._errors.remove(NON_FIELD_ERRORS)
    this._errors.removeAll(fields)
  }
}

// ================================================================= Changes ===

/**
 * Determines which fields have changed from initial form data.
 * @param {boolean=} _hasChangedCheck if true, the method is only being run to
 *   determine if any fields have changed, not to get the list of fields.
 * @return {Array.<string>|boolean} a list of changed field names or true if
 *   only checking for changes and one is found.
 */
Form.prototype.changedData = function(_hasChangedCheck) {
  var changedData = []
  var initialValue
  // XXX: For now we're asking the individual fields whether or not
  // the data has changed. It would probably be more efficient to hash
  // the initial data, store it in a hidden field, and compare a hash
  // of the submitted data, but we'd need a way to easily get the
  // string value for a given field. Right now, that logic is embedded
  // in the render method of each field's widget.
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    var name = fieldNames[i]
    var field = this.fields[name]
    var prefixedName = this.addPrefix(name)
    var dataValue = field.widget.valueFromData(this.data, this.files, prefixedName)
    if (!field.showHiddenInitial) {
      initialValue = object.get(this.initial, name, field.initial)
      if (is.Function(initialValue)) {
        initialValue = initialValue()
      }
    }
    else {
      var initialPrefixedName = this.addInitialPrefix(name)
      var hiddenWidget = new field.hiddenWidget()
      try {
        initialValue = hiddenWidget.valueFromData(
                this.data, this.files, initialPrefixedName)
      }
      catch (e) {
        if (!(e instanceof ValidationError)) { throw e }
        // Always assume data has changed if validation fails
        if (_hasChangedCheck) {
          return true
        }
        changedData.push(name)
        continue
      }
    }
    if (field._hasChanged(initialValue, dataValue)) {
      if (_hasChangedCheck) {
        return true
      }
      changedData.push(name)
    }
  }
  if (_hasChangedCheck) {
    return false
  }
  return changedData
}

/**
 * @return {boolean} true if input data differs from initial data.
 */
Form.prototype.hasChanged = function() {
  this._lastHasChanged = this.changedData(true)
  return this._lastHasChanged
}

// ================================================================== Status ===

/**
 * @return {boolean} true if the form needs a callback argument for final
 *   validation.
 */
Form.prototype.isAsync = function() {
  if (this.clean.length == 1) { return true }
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    var customClean = this._getCustomClean(fieldNames[i])
    if (is.Function(customClean) && customClean.length == 1) {
      return true
    }
  }
  return false
}

/**
 * @return {boolean} true if all required fields have been completed.
 */
Form.prototype.isComplete = function() {
  if (!this.isValid() || this.isPending()) {
    return false
  }
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l; i++) {
    var fieldName = fieldNames[i]
    if (this.fields[fieldName].required &&
        typeof this.cleanedData[fieldName] == 'undefined') {
      return false
    }
  }
  return true
}

/**
 * @return {boolean} true if the form needs to be multipart-encoded, in other
 *   words, if it has a FileField.
 */
Form.prototype.isMultipart = function() {
  var fieldNames = Object.keys(this.fields)
  for (var i = 0, l = fieldNames.length; i < l ; i++) {
    if (this.fields[fieldNames[i]].widget.needsMultipartForm) {
      return true
    }
  }
  return false
}

/**
 * @return {boolean} true if the form is waiting for async validation to
 *   complete.
 */
Form.prototype.isPending = function() {
  return !is.Empty(this._pendingAsyncValidation)
}

/**
 * @return {boolean} true if the form doesn't have any errors.
 */
Form.prototype.isValid = function() {
  if (this.isInitialRender) {
    return false
  }
  return !this.errors().isPopulated()
}

/**
 * @return {boolean} true if the form is waiting for async validation of its
 *   clean() method to complete.
 */
Form.prototype.nonFieldPending = function() {
  return typeof this._pendingAsyncValidation[NON_FIELD_ERRORS] != 'undefined'
}

/**
 * @return {boolean} true if this form is allowed to be empty and if input data
 *   differs from initial data. This can be used to determine when required
 *   fields in an extra FormSet form become truly required.
 */
Form.prototype.notEmpty = function() {
  return (this.emptyPermitted && this._lastHasChanged === true)
}

// ================================================================ Prefixes ===

/**
 * Adds an initial prefix for checking dynamic initial values.
 * @param {string} fieldName a field name.
 * @return {string}
 */
Form.prototype.addInitialPrefix = function(fieldName) {
  return 'initial-' + this.addPrefix(fieldName)
}

/**
 * Prepends a prefix to a field name if this form has one set.
 * @param {string} fieldName a form field name.
 * @return {string} the field name with a prefix prepended if this form has a
 *   prefix set, otherwise the field name as-is.
 * @return {string}
 */
Form.prototype.addPrefix = function(fieldName) {
  if (this.prefix !== null) {
    return formatObj(this.prefixFormat, {prefix: this.prefix, name: fieldName})
  }
  return fieldName
}

/**
 * Extracts a form field name from the name attribute of a rendered form input.
 * @param {string} htmlName a name attribute from a rendered form input.
 * @return {string}
 */
Form.prototype.removePrefix = function(htmlName) {
  if (this.prefix !== null) {
    var partialPrefix = this.prefixFormat.replace('{prefix}', this.prefix)
    var startIndex = partialPrefix.indexOf('{name}')
    var lengthDiff = htmlName.length - partialPrefix.length
    return htmlName.substr(startIndex, /* length of {name} */ 6 + lengthDiff)
  }
  return htmlName
}

/**
 * Creates a version of the given data object with prefixes removed from the
 * property names if this form has a prefix, otherwise returns the object
 * itself.
 * @param {object.<string,*>} data
 * @return {Object.<string,*>}
 */
Form.prototype._deprefixData = function(data) {
  if (this.prefix == null) { return data }
  var prefixedData = {}
  var fieldNames = Object.keys(data)
  for (var i = 0, l = fieldNames.length; i < l; i++) {
    prefixedData[this.removePrefix(fieldNames[i])] = data[fieldNames[i]]
  }
  return prefixedData
}

/**
 * Creates a version of the given data object with prefixes added to the
 * property names if this form has a prefix, otherwise returns the object
 * itself.
 * @param {object.<string,*>} data
 * @return {Object.<string,*>}
 */
Form.prototype._prefixData = function(data) {
  if (this.prefix == null) { return data }
  var prefixedData = {}
  var fieldNames = Object.keys(data)
  for (var i = 0, l = fieldNames.length; i < l; i++) {
    prefixedData[this.addPrefix(fieldNames[i])] = data[fieldNames[i]]
  }
  return prefixedData
}

module.exports = Form